/* fade.s */
/***********************************************************************

                             fade()

************************************************************************
This star models multipath fading channels for mobile radio applications.
The star accepts a complex baseband equivalent input and produces complex 
baseband equivalent samples. The method is based on William C. Jakes, 
"Microwave Mobile Communications," John Wiley & Sons, 1974 in particular 
pp. 13-65.  Multipath is modeled using the technique presented by Nader 
Farahati, "A Software Multipath Fading Channel Simulator", Technophone 
Limited, July 1989. Nader Farahati is now with Scientific Generics, Cambridge
U.K. Each multipath is associated with a time delay. The time delays are 
incorporated by transforming the problem into the frequency domain.

This star first reads all samples, u(t), at its input. It then multiplies the
complex input samples by the complex fading channel amplitude with Doppler shift
for path i, ri(t),  and transforms them into the frequency domain, Yi(f). 
The various delays are incorporated by multiplying the frequency domain data, 
Yi(f), by exp{-2PIj(fc+f)ti} where fc is the carrier frequency, f is the
frequency, and ti is the time delay of the ith multipath.

The various multipaths with independent fading channel amplitudes are
added in the frequency domain and transformed back into the time domain.

The star then outputs the complex data as two channels (in-phase and
quad-phase) in 128 sample chunks. This helps in limiting the size of
buffers. 

Note that other doppler spectrums and Rician distributions will be
supported later. 

Programmer: 	Sasan Ardalan	
Date: 		Dec. 27, 1990
Modified:       Jeyhan Karaoguz 
Date:           June 28, 1991

*/

#include <math.h>
#define PI 3.1415926

parameters
	param_def = "Number of points (preferably a power of 2)";	
	int npts = 128;
	param_def = "Doppler Spectrum, only Ez supported at this time.";
	int	type=0;
	param_def = "Sampling Rate";	
	float	fs=1.0;
	param_def = "Carrier frequency";	
	float	fc=1000e6;
	param_def = "Vehicle Velocity, m/s";	
	float	v= 0;
	param_def = "Power";	
	float	p= 1.0;
	param_def= "Array of multipath delays microsec: number_of_paths t0 t1  ... ";
        array delays;
	param_def= "Array of multipath powers: number_of_paths p0 p1  ... ";
        array powers;
	param_def= "Number of Plane Waves arriving plane waves, N where N >=34";
	int	numberArrivals=40; 
end

states
	int	inbufs;
	int fftl;
	int fftexp;
	float* fftBuffer_P;
	float* inputBuffer_P;
	float* outputBuffer_P;
	int inSampleCount;
	int samplesOutput;
	float wd;
end


output_buffer
	float inPhase;
	float quadPhase;
end

declarations
	int numberOfSamples;
	int i,j;
	float fd, lambda;
	double er,ei;
	float arg;
	float t;
	float env;
	float tsin,tcos;
	int	samples;
end

initialization_code

	inbufs = no_input_buffers();
	/*
 	 * extract the exponent in npts=2**fftexp 
         */
        fftexp = (int) (log((float)npts)/log(2.0)+0.5);
        fftl = 1 << fftexp;
        if (fftl > npts ) {
                    fftl = fftl/2;
                    fftexp -= 1;
        }


	if (n_delays == 0 )
	{
		fprintf(stderr,"fade: number of delays must be >= 1 \n");
		return(1);
	}
	if (n_delays != n_powers)
	{
		fprintf(stderr,"fade: number of delays dont match powers \n");
		return(2);
	}

	if ((fftBuffer_P = (float*)calloc(2*fftl,sizeof(float))) == NULL)
	{
		fprintf(stderr,"fade: can't allocate work space \n");
		return(3);
	}
	if ((inputBuffer_P = (float*)calloc(2*fftl,sizeof(float))) == NULL)
	{
		fprintf(stderr,"fade: can't allocate work space \n");
		return(4);
	}
	if ((outputBuffer_P = (float*)calloc(2*fftl,sizeof(float))) == NULL)
	{
		fprintf(stderr,"fade: can't allocate work space \n");
		return(5);
	}
	for (i=0;i<(2*fftl);i++) 
	{
		fftBuffer_P[i] = 0.0;
		inputBuffer_P[i] = 0.0;
		outputBuffer_P[i] = 0.0;
	}
	/*
	 * calculate maximum doppler shift
 	 */
	lambda= 3.e8/fc;
	fd= v/lambda;
	wd=2*PI*fd;

	inSampleCount=0;
	samplesOutput=0;
end

main_code
/*
 * input the samples. Collect fftl number of samples before processing
 */
for (numberOfSamples = min_avail(); numberOfSamples > 0; --numberOfSamples) {

      it_in(0);
      it_in(1);
      /*
       * store samples in inputBuffer_P
       */
      inputBuffer_P[2*inSampleCount] = inf(0,0);
      inputBuffer_P[2*inSampleCount+1] = inf(1,0);
      /*
       * keep track of the number of samples inputted up to this point
       */
      inSampleCount++;
      if(inSampleCount == fftl) {
	/*
	 * fftl number of samples collected. 
	 * Compute fading channel amplitude. 
	 * Compute n_delays, the number of multipaths, independent amplitudes.
	 * See Jakes(1974).
	 */
        for(j=0; j< n_delays; j++) {
	   for(samples=0; samples < inSampleCount; samples++) {
	      t=(float)samples/fs;
	      /*
	       * Compute complex mobile fading channel amplitude
	       */
	      er=0.0;
	      ei=0.0;
	      for(i=0; i<numberArrivals/2; i++) {
		 /*
		  * note the term (i+2*j)
		  * This poduces (hopefully) independent amplitues for
		  * different j's where j is the multipath number.
		  * ri(t) = er + j*ei
		  */
	         arg = wd*t*cos(2*PI*(float)i/numberArrivals) + PI*((float)(i+2*j)/numberArrivals);
	         er=er + cos(arg);
	         ei=ei + sin(arg);
	      }
	      /*
	       * powers[j] are the relative powers for each multipath
	       */	
	      er= sqrt(2.0*p*powers[j]/numberArrivals) * er;
	      ei = sqrt(2.0*p*powers[j]/numberArrivals) * ei;
	      /*
	       * form u(t)*ri(t). Store in fft buffer to be transformed
	       * into the frequency domain.
	       * note  that u(t) is stored in inputBuffer_P.
	       */
	      fftBuffer_P[2*samples] = er*inputBuffer_P[2*samples] - ei*inputBuffer_P[2*samples+1]; 
	      fftBuffer_P[2*samples+1] = ei*inputBuffer_P[2*samples] + er*inputBuffer_P[2*samples+1]; 
	   }
	   /*
	    * Transform into the frequency domain
	    */
	    cxfft(fftBuffer_P,&fftexp); 
	      
	   /*
	    * incorporate delay by multiplying Yi(f) by
	    * exp { -j2Pi(fc +f)ti} where the ti are stored in
	    * delays[i]. Also add all paths.
	    */
	   for(samples=0; samples < inSampleCount; samples++) {
	       arg = 2*PI*(fc+(float)samples*fs/(float)inSampleCount)*1.e-6*delays[j];
	       tcos = cos(arg);
	       tsin = sin(arg);
	    
	       outputBuffer_P[2*samples] += (tcos*fftBuffer_P[2*samples] - tsin*fftBuffer_P[2*samples+1]);
	       outputBuffer_P[2*samples+1] += (tsin*fftBuffer_P[2*samples] + tcos*fftBuffer_P[2*samples+1]);
	   }
	}
	/*
	 * convert to time domain
	 */
	 cxifft(outputBuffer_P,&fftexp); 

         while(samplesOutput < fftl) { 
         	for(i=0; i < 128; i++) {
	        	samples=samplesOutput;
	    		it_out(0);
	    		it_out(1);
	    		inPhase(0) = outputBuffer_P[2*samples]/fftl;
	    		quadPhase(0) = outputBuffer_P[2*samples+1]/fftl;
	    		samplesOutput++;
     		 } 
	  }

      }
}
return(0);
end

wrapup_code
	/* 
	 * free up allocated space	
	 */
	free((char*)fftBuffer_P);
	free((char*)inputBuffer_P);
	free((char*)outputBuffer_P);
end

