/*
 * standard.h
 *
 * CCSP Program Header.  This file contains macro definitions to be used
 * throughout CCSP programs.  Most of these definitions are used for
 * debugging and profiling purposes.
 *
 * (following note from joseph hall, 6/24/87)
 *
 * If anyone's curious, these macros derive largely from the work of a
 * friend of mine, Dean Thompson, at Carnegie-Mellon.  He spent a couple
 * of months (at least) putting together these macros, some debugging code
 * and a nice set of coding standards for the Camelot project there.
 * Hopefully I can pirate enough of his efforts to save us some trouble,
 * though I'm sure I'll also be forced to display some originality as the
 * project progresses.
 *
 * (following note from joseph hall, 8/6/87)
 *
 */


#define notrace	0
#define nodebug	0

/*
 * DIAGNOSTIC REPORTING MACROS
 *
 * These use the "message machine" interprocess communication routines in
 * mm_mmach.  That way diagnostic messages can be send to another running
 * process that displays and/or stores them interactively.
 *
 * The current implementation of the message machine uses QIOW calls that
 * wait for I/O completion before returning.  This means that if debugging
 * is "on" (i. e., progDebugLevel > 0) then there must be some process running
 * that will read the messages sent to the message machine mailbox by these
 * macros, else the program will hang while it waits for its message to get
 * picked up.
 */

extern void	Mm_Init();
extern void	Mm_SendMsg();
extern void	Mm_SendEOF();

extern int	_debugLevel;
extern int	_recursionLevel;
extern int	_blockNestLevel;
extern int	_traceOn;

/*
 * Call DEBUG_INIT once and only once before using any of the other
 * DEBUG macros.
 */


#ifdef nodebug

#define DEBUG_INIT
#define DEBUG(n, string)
#define DEBUGF(n, sprintf_args)
#define DEBUG_OFF

#else nodebug

#define	DEBUG_INIT		Mm_Init()


/*
 * There are two flavors of the message macro here.
 *
 * Flavor A, DEBUG(n, string), takes an integer argument n, which we agree
 * will be in the range 0 to 100 inclusive, and a pointer to a string.  It
 * writes the string to the mailbox if the current "debug level" of the
 * program is greater than or equal to n.
 *
 * Flavor B, DEBUGF(n, sprintf_args) takes an integer argument 0 to 100
 * inclusive and a complete, parenthesized sprintf argument list.  The
 * result of the sprintf MUST be a buffer named msg.  To wit:
 *
 *   DEBUGF(25, (msg, "This is an example of a DEBUGF macro.  N = %d.\n",
 *               n));
 */

#define DEBUG(n, string)						\
{									\
	if (_debugLevel >= (n)) {					\
		if (_traceOn) {						\
			char	_buffer[MM_MAILBOX_SIZE];		\
			int	i;					\
									\
			for (i = 0; i < _recursionLevel + 		\
					_blockNestLevel + 3; i++) {	\
				_buffer[i] = ' ';			\
			}						\
									\
			_buffer[i] = 0;					\
			(void) strcat(_buffer, string);			\
			Mm_SendMsg(_buffer);				\
		}							\
		else {							\
			Mm_SendMsg(string);				\
		}							\
	}								\
}

#define DEBUGF(n, sprintf_args)						\
{									\
	if (_debugLevel >= (n)) {					\
		char 	msg[MM_MAILBOX_SIZE] = "";			\
									\
		(void) sprintf sprintf_args;				\
									\
		if (_traceOn) {						\
			char	_buffer[256];				\
			int	i;					\
									\
			for (i = 0; i < _recursionLevel +		\
					_blockNestLevel + 3; i++) {	\
				_buffer[i] = ' ';			\
			}						\
									\
			_buffer[i] = 0;					\
			(void) strcat(_buffer, msg);			\
			Mm_SendMsg(_buffer);				\
		}							\
		else {							\
			Mm_SendMsg(msg);				\
		}							\
	}								\
}


/*
 * DEBUG_OFF sends an EOF signal to the mailbox.  This will cause the
 * process reading the mailbox to exit, or at least that's the intent.
 * This should be called once and only once after you're done using the
 * DEBUG macros.
 */

#define DEBUG_OFF		Mm_SendEOF()

#endif nodebug

/*
 * DEBUGGING/PROFILING HOOKS
 *
 * All functions will begin with BEGIN("routine-name") and end with END
 * instead of braces.  This will enable us to insert tracing and/or
 * profiling code later on.  For now these macros expand into braces
 * anyway.
 *
 * Someone wanting to add tracing hooks to loops or blocks inside his
 * routines may also use BLOCK("block-name") and ENDBLOCK instead of
 * left and right braces.
 *
 * RETURN(argument) and RET are used to terminate functions and procedures,
 * which must ALWAYS exit through them, obviously, or they won't be effective.
 * It's bad style, besides, to allow routines to exit by "falling out
 * the bottom."  Using RET and RETURN(argument) has the side-effect of
 * making this particular kind of bad style obvious when it occurs.
 *
 * Example:
 *
 * PUBLIC int Proc1(x)
 *     int i, x, intResult;
 * BEGIN("Proc1")
 *     ....
 *     for (i == 1, i < VALUE, i++) BLOCK("Proc1-loop")
 *         ....
 *         if (i == QUIT_NOW){
 *             ....
 *             RETURN(intResult);
 *         }
 *     ENDBLOCK;
 *     ....
 *     RETURN(intResult);
 * END;
 */

#ifdef notrace

#define BEGIN(routine_name)	{
#define END			}
#define BLOCK(block_name)	{
#define ENDBLOCK		}
#define RET			return
#define RETURN(value)		return(value)

#else notrace

#define BEGIN(routine_name)						\
{									\
	static char	*_routineName = routine_name;			\
	char		_buffer[256];					\
	int		_oldBlockNestLevel;				\
									\
	_recursionLevel++;						\
	_oldBlockNestLevel = _blockNestLevel;				\
	_blockNestLevel = 0;						\
	if (_traceOn) {							\
		int i;							\
		for (i = 0; i < _recursionLevel; i++) {			\
			_buffer[i] = ' ';				\
		}							\
		_buffer[i] = 0;						\
		(void) strcat(_buffer, ">> ");				\
		(void) strcat(_buffer, _routineName);			\
		(void) strcat(_buffer, "\n");				\
		Mm_SendMsg(_buffer);					\
	}								\
	{



#define	END								\
	}								\
}


#define BLOCK(block_name)						\
{									\
	static char	*_blockName = block_name;			\
	char		_buffer[256];					\
									\
	_blockNestLevel++;						\
	if (_traceOn) {							\
		int i, j;						\
		for (i = 0; i < _recursionLevel; i++) {			\
			_buffer[i] = ' ';				\
		}							\
		_buffer[i++] = '>';					\
		for (j = 0; j < _blockNestLevel; i++, j++) {		\
			_buffer[i] = ' ';				\
		}							\
		_buffer[i++] = '>';					\
		_buffer[i++] = ' ';					\
		_buffer[i] = 0;						\
		(void) strcat(_buffer, _blockName);			\
		(void) strcat(_buffer, "\n");				\
		Mm_SendMsg(_buffer);					\
	}								\
	{

#define ENDBLOCK							\
	}								\
	{								\
		char		_buffer[256];				\
									\
		if (_traceOn) {						\
			int i, j;					\
			for (i = 0; i < _recursionLevel; i++) {		\
				_buffer[i] = ' ';			\
			}						\
			_buffer[i++] = '<';				\
			for (j = 0; j < _blockNestLevel; i++, j++) {	\
				_buffer[i] = ' ';			\
			}						\
			_buffer[i++] = '<';				\
			_buffer[i++]= ' ';				\
			_buffer[i] = 0;					\
			(void) strcat(_buffer, _blockName);		\
			(void) strcat(_buffer, "\n");			\
			Mm_SendMsg(_buffer);				\
		}							\
		_blockNestLevel--;					\
	}								\
}


#define RET								\
	{								\
		char		_buffer[256];				\
									\
		if (_traceOn) {						\
			int i;						\
			for (i = 0; i < _recursionLevel; i++) {		\
				_buffer[i] = ' ';			\
			}						\
			_buffer[i] = 0;					\
			(void) strcat(_buffer, "<< ");			\
			(void) strcat(_buffer, _routineName);		\
			(void) strcat(_buffer, "\n");			\
			Mm_SendMsg(_buffer);				\
		}							\
		_recursionLevel--;					\
		_blockNestLevel = _oldBlockNestLevel;			\
		return;							\
	}


#define RETURN(value)							\
	{								\
		char		_buffer[256];				\
									\
		if (_traceOn) {						\
			int i;						\
			for (i = 0; i < _recursionLevel; i++) {		\
				_buffer[i] = ' ';			\
			}						\
			_buffer[i] = 0;					\
			(void) strcat(_buffer, "<< ");			\
			(void) strcat(_buffer, _routineName);		\
			(void) strcat(_buffer, "\n");			\
			Mm_SendMsg(_buffer);				\
		}							\
		_recursionLevel--;					\
		_blockNestLevel = _oldBlockNestLevel;			\
		return(value);						\
	}


#endif notrace

/*
 * ARGUMENT LABELS
 *
 * These are used to label function arguments.
 *
 * IN indicates that the argument data is read only by the function, and 
 * not written.  If the function treats ALL argument data like this, then
 * the IN declaration may be omitted.
 *
 * INOUT indicates that the argument data is read and written by the
 * function.
 *
 * OUT indicates that the argument data is written by the function,
 * without reference to its previous value.
 *
 * INOUT and OUT apply only to pointer arguments, since there is no
 * direct pass-by-reference mechanism in C.
 *
 * A fourth keyword, SHARED, is also defined.  It indicates that the
 * argument is a pointer to shared data (i. e., shared between PROCESSES
 * or simultaneously-running coroutines).  Data can be read or written,
 * but not meaningfully copied, since the value of the data pointed to
 * may CHANGE at any time.  A good example would be an address containing
 * a time.  If this seems niggling or excessively complex, well, it's
 * doubtful we'll have use for it anyway.  (But keep it in mind.)
 *
 * Example:
 *
 * char CharFunc(    IN readOnlyArg,
 *		    OUT writeOnlyArg, writeOnlyArg2,
 *		  INOUT readWriteArg,
 *		 SHARED theTime)
 */

#define	IN
#define OUT
#define INOUT
#define SHARED



/*
 * FUNCTION SCOPE LABELS
 *
 * These are intended to show the scope of a function's use.
 *
 * The keyword GLOBAL indicates that the function will be used THROUGHOUT
 * THE ENTIRE SYSTEM.  These should be rare, and, for the most part,
 * lumped together (but sensibly) in a library of utilities.
 *
 * The keyword PUBLIC indicates that the function will be used only
 * within its "home" MODULE.  Note that a module will often comprise
 * several files, thus the distinction between PUBLIC and . . .
 *
 * . . . PRIVATE, which indicates that the function will be used only
 * within its home FILE.  PRIVATE expands to "static," so this is
 * enforced.
 *
 * Note that if a module is small enough to fit in one file it will use
 * only PRIVATE and GLOBAL labels.
 */

#define GLOBAL
#define PUBLIC
#define PRIVATE static

#ifdef MIN
#undef MIN
#endif
#ifdef MAX
#undef MAX
#endif

#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))


